{**************************************}
{* Main unit for Obstquiz application *}
{**************************************}

unit quiz;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ExtCtrls, Grids, Menus, LazUTF8;

const
  iTotalFruits = 50;
  iGroup1Fruits = 15;
  iGroup2Fruits = 30;
  iGroup3Fruits = 50;

type
  TFruit = record
    Name, Name2, Name3, Shortname: string;
    Pics, Group: Integer;
  end;
  TFruits = array[0..iTotalFruits - 1] of TFruit;
  {************}
  { TfObstquiz }
  {************}
  TfObstquiz = class(TForm)
    MainMenu1: TMainMenu;
    mQuiz, mQuizNew, mQuizExit: TMenuItem;
    mOptions, mOptionsFruits, mOptionsFruits1, mOptionsFruits2, mOptionsFruits3: TMenuItem;
    mHelp, mHelpHelp, mHelpAbout: TMenuItem;
    StaticText1: TStaticText;
    laFruit: TLabel;
    imFruit1, imFruit2, imFruit3, imFruit4: TImage;
    imFruit5, imFruit6, imFruit7, imFruit8: TImage;
    Shape1, Shape2, Shape3, Shape4: TShape;
    Shape5, Shape6, Shape7, Shape8: TShape;
    shSel1, shSel2, shSel3, shSel4: TShape;
    Label2, laFruits: TLabel;
    sgEval: TStringGrid;
    btStart: TButton;
    procedure FormCreate(Sender: TObject);
    procedure mQuizNewClick(Sender: TObject);
    procedure mQuizExitClick(Sender: TObject);
    procedure mOptionsFruits1Click(Sender: TObject);
    procedure mOptionsFruits2Click(Sender: TObject);
    procedure mOptionsFruits3Click(Sender: TObject);
    procedure mHelpHelpClick(Sender: TObject);
    procedure mHelpAboutClick(Sender: TObject);
    procedure btStartClick(Sender: TObject);
    procedure imFruit1Click(Sender: TObject);
    procedure imFruit2Click(Sender: TObject);
    procedure imFruit3Click(Sender: TObject);
    procedure imFruit4Click(Sender: TObject);
    procedure imFruit5Click(Sender: TObject);
    procedure imFruit6Click(Sender: TObject);
    procedure imFruit7Click(Sender: TObject);
    procedure imFruit8Click(Sender: TObject);
  private
    iGroup0, iGroup, iFruits, iFruit, iCorrect, iPic: Integer;
    sFruit: string;
    bAnswer: Boolean;
    aFruits: TFruits;
    aFruitsDone: array[0..iTotalFruits - 1] of Boolean;
    imFruits: array[0..7] of TImage;
    shSelection: array[0..3] of TShape;
  end;

var
  fObstquiz: TfObstquiz;

implementation

{$R *.lfm}

{ Format number for the grid (right-align) }

function GFormat(N: Integer; S: string): string;

var
  SN: string;

begin
  SN := ' ' + IntToStr(N);
  if N < 10 then
    SN := '  ' + SN
  else if N < 100 then
    SN := ' ' + SN;
  if S = '' then
    SN := ' ' + SN
  else
    SN := SN + S;                                                              // % sign
  Result := SN;
end;

{ Read fruit data from text file }

procedure ReadFruits(out Fruits: TFruits);

var
  N, P: Integer;
  Filename, Line, S: string;
  InFile: Text;

begin
  Filename := GetCurrentDir + '/data/obst.txt'; DoDirSeparators(Filename);
  Assign(InFile, Filename); Reset(InFile); N := 0;
  while not EoF(InFile) do begin
    Readln(InFile, Line); Line := UTF8Trim(Line);
    if Line <> '' then begin
      Inc(N);
      if N <= iTotalFruits then
        with Fruits[N - 1] do begin
          Name := UTF8Trim(UTF8Copy(Line, 1, 15));                             // cols 1-15: fruit name
          Name2 := ''; Name3 := ''; ShortName := '';
          if Length(Line) > 20 then begin                                      // cols 20-: supplementary names
            // Read second (and maybe third) name or short name
            S := UTF8Trim(UTF8Copy(Line, 20, UTF8Length(Line)));
            P := UTF8Pos(',', S);                                              // second and third name separated by comma
            if P = 0 then begin
              // One single name given
              if LeftStr(S, 1) = '*' then begin
                // Name given is shortname
                ShortName := S;
                UTF8Delete(ShortName, 1, 1);
              end
              else begin
                // Name given is second name
                Name2 := S;
              end;
            end
            else begin
              // Two names given
              Name2 := UTF8Copy(S, 1, P - 1);
              Name3 := UTF8Copy(S, P + 1, UTF8Length(S));
            end;
          end;
          Pics := StrToInt(UTF8Copy(Line, 16, 1));                             // col 16: number of fruit pictures
          Group := StrToInt(UTF8Copy(Line, 18, 1));                            // col 18: fruit group indicator
        end;
    end;
  end;
  Close(InFile);
end;

{ Select fruit image clicked on as user answer }

procedure FruitSelect(PicClicked, PicCorrect: Integer; var MayAnswer: Boolean; var FruitsIm: array of TImage;
  var SelShapes: array of TShape; Fruits, GroupFruits: Integer; var Correct: Integer);

var
  I: Integer;
  Colour: TColor;

begin
  if MayAnswer then begin
    // Proceed only if an answer is to be given (i.e. after a new question has been asked)
    if PicClicked = PicCorrect then begin
      // Correct picture clicked: Draw green border
      Colour := clLime;
      Inc(Correct);
    end
    else begin
      // Wrong picture clicked: Draw red border
      Colour := clRed;
    end;
    // Color the border and make it visible around picture clicked on
    SelShapes[0].Left := FruitsIm[PicClicked].Left; SelShapes[0].Top := FruitsIm[PicClicked].Top;
    SelShapes[1].Left := FruitsIm[PicClicked].Left; SelShapes[1].Top := FruitsIm[PicClicked].Top + FruitsIm[PicClicked].Height - 5;
    SelShapes[2].Left := FruitsIm[PicClicked].Left; SelShapes[2].Top := FruitsIm[PicClicked].Top;
    SelShapes[3].Left := FruitsIm[PicClicked].Left + FruitsIm[PicClicked].Width - 5; SelShapes[3].Top := FruitsIm[PicClicked].Top;
    for I := 0 to 3 do begin
      SelShapes[I].Brush.Color := Colour;
      SelShapes[I].Visible := True;
    end;
    // Fill in evaluation grid
    fObstquiz.sgEval.Cells[1, 0] := GFormat(Fruits, '');
    fObstquiz.sgEval.Cells[1, 1] := GFormat(Correct, '');
    fObstquiz.sgEval.Cells[1, 2] := GFormat(Fruits - Correct, '');
    fObstquiz.sgEval.Cells[1, 3] := GFormat(Round(100 * (Correct / Fruits)), '%');
    if Fruits < GroupFruits then begin
      // Still fruits left: Enable the "Frage" (= question) button
      fObstQuiz.btStart.Enabled := True; MayAnswer := False;
    end
    else begin
      // All fruits done: End of quiz message
      fObstQuiz.btStart.Enabled := False; MayAnswer := False;
      MessageDlg('Obstquiz', 'Alle Früchte dieser Gruppe wurden gezeigt. Ende des Quizes.', mtInformation, [mbOK], 0);
    end;
  end;
end;

{************}
{ TfObstquiz }
{************}

{ Application start: Initialisation }

procedure TfObstquiz.FormCreate(Sender: TObject);

begin
  // Create arrays with fruit images and selection-border shapes
  imFruits[0] := imFruit1; imFruits[1] := imFruit2; imFruits[2] := imFruit3; imFruits[3] := imFruit4;
  imFruits[4] := imFruit5; imFruits[5] := imFruit6; imFruits[6] := imFruit7; imFruits[7] := imFruit8;
  shSelection[0] := shSel1; shSelection[1] := shSel2; shSelection[2] := shSel3; shSelection[3] := shSel4;
  // Read fruits data from file
  ReadFruits(aFruits);
  // Start a new quiz (with group 1 fruits)
  Randomize;
  iGroup0 := 1;
  mQuizNew.Click;
end;

{ Menu item "Quiz > Neu": Start a new quiz }

procedure TfObstquiz.mQuizNewClick(Sender: TObject);

var
  I: Integer;

begin
  iGroup := iGroup0;                                                           // user selection becomes active now
  bAnswer := False;                                                            // disable action when images are clicked
  laFruit.Caption := 'Welches dieser Bilder zeigt ...?';
  // Total number of fruits (questions) for actual quiz
  case iGroup of
    1: iFruits := iGroup1Fruits;
    2: iFruits := iGroup2Fruits;
    3: iFruits := iGroup3Fruits;
  end;
  // Variables initialisation and form controls reset
  iFruit := 0; iCorrect := 0;
  for I := 0 to iTotalFruits - 1 do
    aFruitsDone[I] := False;
  for I := 0 to 7 do
    imFruits[I].Picture.Clear;
  for I := 0 to 3 do
    shSelection[I].Visible := False;
  laFruits.Visible := False;
  for I := 0 to 3 do
    sgEval.Cells[1, I] := '';
  // Enable "Start" button (-> user can make generate a new question)
  btStart.Enabled := True;
end;

{ Menu item "Quiz > Verlassen": Exit application }

procedure TfObstquiz.mQuizExitClick(Sender: TObject);

begin
  Close;
end;

{ Menu items "Einstellungen > Früchte > ...": Fruit group (number of fruits) selection }

procedure TfObstquiz.mOptionsFruits1Click(Sender: TObject);

begin
  mOptionsFruits1.Checked := True; mOptionsFruits2.Checked := False; mOptionsFruits3.Checked := False;
  iGroup0 := 1;
end;

procedure TfObstquiz.mOptionsFruits2Click(Sender: TObject);

begin
  mOptionsFruits1.Checked := False; mOptionsFruits2.Checked := True; mOptionsFruits3.Checked := False;
  iGroup0 := 2;
end;

procedure TfObstquiz.mOptionsFruits3Click(Sender: TObject);

begin
  mOptionsFruits1.Checked := False; mOptionsFruits2.Checked := False; mOptionsFruits3.Checked := True;
  iGroup0 := 3;
end;

{ Menu item "Hilfe > Hilfe": Display (short) application help text }

procedure TfObstquiz.mHelpHelpClick(Sender: TObject);

var
  S: string;

begin
  S := 'Mit dem Knopf "Frage" wird nach einer neuen Frucht gefragt. Zum Antworten, auf das Bild, das diese Fucht zeigt, ';
  S += 'klicken. Ein grüner Rahmen signalisiert eine richtige, ein roter Rahmen eine falsche Antwort.' + LineEnding;
  S += 'Im Menü "Einstellungen" kann die Fruchtgruppe (Anzahl der Früchte/Fragen) ausgewählt werden. Mit "Neu" im Menü ';
  S += '"Quiz" wird ein neuer Quiz gestartet.';
  MessageDlg('"Obstquiz" Hilfe', S, mtInformation, [mbOK], 0);
end;

{ Menu item "Hilfe > Über": Display application about }

procedure TfObstquiz.mHelpAboutClick(Sender: TObject);

var
  S: string;

begin
  S := 'Obstquiz für Kinder:' + LineEnding;
  S += 'Erraten, auf welchem der 8 angezeigten Bilder die vorgegebene Frucht zu sehen ist.' + LineEnding + LineEnding;
  S += 'Version 1.0.1, © allu, Oktober-Dezember 2021.';
  MessageDlg('Über "Obstquiz"', S, mtInformation, [mbOK], 0);
end;

{ Button "Frage" pushed: Generate new question (ask for new fruit) }

procedure TfObstquiz.btStartClick(Sender: TObject);

var
  F, F0, R, I, J: Integer;
  Filename, S: string;
  OK: Boolean;
  FShown: array[0..7] of Integer;

begin
  Inc(iFruit);
  laFruits.Caption := 'Frucht ' + IntToStr(iFruit) + '/' + IntToStr(iFruits); laFruits.Visible := True;
  for I := 0 to 3 do
    shSelection[I].Visible := False;
  // Random fruit (within actual group and not yet asked)
  repeat
    F0 := Random(iTotalFruits);
  until (aFruits[F0].Group <= iGroup) and (not aFruitsDone[F0]);
  aFruitsDone[F0] := True;                                                     // mark this fruit as "done"
  sFruit := aFruits[F0].Name;
  // Randomly choose one of the possibly 3 fruit names (for display in question text)
  R := 0;
  if (aFruits[F0].Name2 <> '') and (aFruits[F0].Name3 <> '') then
    R := Random(3)
  else if aFruits[F0].Name2 <> '' then
    R := Random(2);
  case R of
    0: S := sFruit;
    1: S := aFruits[F0].Name2;
    2: S := aFruits[F0].Name3;
  end;
  laFruit.Caption := 'Welches dieser Bilder zeigt die Frucht ' + UTF8UpperCase(S) + '?';
  // Randomly choose 7 fruit images
  for I := 0 to 7 do begin
    FShown[I] := 0;
    repeat
      OK := False;
      F := Random(iTotalFruits);
      if (F <> F0) and (aFruits[F].Group <= iGroup) then begin
        // Fruits must different from the one asked for and be part of actual fruit group
        OK := True;
        if (aFruits[F].Shortname <> '') and (aFruits[F].Shortname = aFruits[F0].Name) then begin
          // Do not use fruits that have a shortname that equals the name of the fruit asked for
          // This avoids displaying for example a watermelon if the question concerns melons
          OK := False;
        end
        else if I > 0 then begin
          // Each fruit should be displayed only once (no duplicates)
          for J := 0 to I - 1 do begin
            if F = FShown[J] then
              OK := False;
          end;
        end;
      end;
    until OK;
    // Display fruit's picture (if there are several pics, randomly choose one of them)
    Filename := GetCurrentDir + '/data/' + UTF8LowerCase(aFruits[F].Name);
    R := Random(aFruits[F].Pics) + 1;
    if R >= 2 then
      Filename += IntToStr(R);
    Filename += '.jpg'; DoDirSeparators(Filename);
    imFruits[I].Picture.LoadFromFile(Filename);
    FShown[I] := F;                                                            // save fruit chosen (for duplicates test)
  end;
  // Now place the fruit asked for at a random position within the picture-grid
  iPic := Random(8);
  Filename := GetCurrentDir + '/data/' + UTF8LowerCase(sFruit);
  R := Random(aFruits[F0].Pics) + 1;
  if R >= 2 then
    Filename += IntToStr(R);
  Filename += '.jpg'; DoDirSeparators(Filename);
  imFruits[iPic].Picture.LoadFromFile(Filename);
  // Disable the "Frage" button and enable action when user clicks a picture (-> users can give their answer)
  bAnswer := True; btStart.Enabled := False;
end;

{ Click on one of the 8 fruit images: Select this fruit as user's answer }

procedure TfObstquiz.imFruit1Click(Sender: TObject);

begin
  FruitSelect(0, iPic, bAnswer, imFruits, shSelection, iFruit, iFruits, iCorrect);
end;

procedure TfObstquiz.imFruit2Click(Sender: TObject);

begin
  FruitSelect(1, iPic, bAnswer, imFruits, shSelection, iFruit, iFruits, iCorrect);
end;

procedure TfObstquiz.imFruit3Click(Sender: TObject);

begin
  FruitSelect(2, iPic, bAnswer, imFruits, shSelection, iFruit, iFruits, iCorrect);
end;

procedure TfObstquiz.imFruit4Click(Sender: TObject);

begin
  FruitSelect(3, iPic, bAnswer, imFruits, shSelection, iFruit, iFruits, iCorrect);
end;

procedure TfObstquiz.imFruit5Click(Sender: TObject);

begin
  FruitSelect(4, iPic, bAnswer, imFruits, shSelection, iFruit, iFruits, iCorrect);
end;

procedure TfObstquiz.imFruit6Click(Sender: TObject);

begin
  FruitSelect(5, iPic, bAnswer, imFruits, shSelection, iFruit, iFruits, iCorrect);
end;

procedure TfObstquiz.imFruit7Click(Sender: TObject);

begin
  FruitSelect(6, iPic, bAnswer, imFruits, shSelection, iFruit, iFruits, iCorrect);
end;

procedure TfObstquiz.imFruit8Click(Sender: TObject);

begin
  FruitSelect(7, iPic, bAnswer, imFruits, shSelection, iFruit, iFruits, iCorrect);
end;

end.

